package container;

import cn.hutool.core.collection.CollUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;

import entity.BaseTunnelInfo;
import enums.TunnelStatus;
import exception.DataExchangeException;
import lombok.extern.slf4j.Slf4j;
import tunnelworker.BaseProtocolTunnel;
import tunnelworker.masters.MasterInterface;
import tunnelworker.slavers.SlaverInterface;

import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;

/**
 * 装载通道对象的容器
 *
 * @author 修唯xiuwei
 * @version 3.0
 */
@Slf4j
public class ProtocolTunnelContainer {

    /**
     * 通道集合
     */
    private Map<String, BaseProtocolTunnel> tunnels = new ConcurrentHashMap<>();


    /**
     * 定时任务集合
     */
    private Map<String, Callable> tasks = new ConcurrentHashMap<>();

    /**
     * 单例构造
     */
    private ProtocolTunnelContainer() {}

    /**
     * 单例
     *
     * @return 实例
     */
    public static final ProtocolTunnelContainer getInstance() {
        return LazyHolder.INSTANCE;
    }

    /**
     * 获取tunnelID下的 tunnel
     *
     * @param tunnelId 通道id
     * @return BaseProtocolTunnel 通道
     */
    public BaseProtocolTunnel getTunnel(String tunnelId) {
        if (this.tunnels.containsKey(tunnelId)) {
            return this.tunnels.get(tunnelId);
        }
        return null;
    }

    /**
     * 获取所有 tunnel 集合
     *
     * @param
     * @return java.util.List<com.jiayue.agcavc.console.tunnelworker.BaseProtocolTunnel>
     * @author L.ym
     * @date 2022/12/6
     **/
    public List<BaseProtocolTunnel> getTunnelList() {
        return CollUtil.newArrayList(this.tunnels.values());
    }

    /**
     * 向集合中置入tunnel
     *
     * @param tunnel 通道对象
     * @return 返回置入的tunnel
     */
    public BaseProtocolTunnel addTunnel(BaseProtocolTunnel tunnel) {
        this.tunnels.put(tunnel.getTunnelId(), tunnel);
        return tunnel;
    }

    /**
     * 删除隧道 同时删除定时更新数据任务
     *
     * @param tunnelId 隧道id
     */
    public void removeTunnel(String tunnelId) {
        this.tunnels.remove(tunnelId);
        stopUpdateTask(tunnelId);
    }

    /**
     * 停止更新任务
     *
     * @param tunnelId 通道id
     */
    public void stopUpdateTask(String tunnelId) {
        if (this.tasks.containsKey(tunnelId)) {
            RecurringTaskContainer.getInstance().removeTask(tasks.get(tunnelId));
            this.tasks.remove(tunnelId);
        }
    }

    /**
     * 添加数据读取任务
     *
     * @param baseProtocolTunnel 通道对象
     */
    public void addUpdateDateTask(BaseProtocolTunnel baseProtocolTunnel) {
        if (this.tasks.containsKey(baseProtocolTunnel.getTunnelId())) {
            RecurringTaskContainer.getInstance().removeTask(this.tasks.get(baseProtocolTunnel.getTunnelId()));
        }
        Callable callable = null;
        if (baseProtocolTunnel instanceof MasterInterface) {
            callable = RecurringTaskContainer.getInstance().addRecurringTask(baseProtocolTunnel.getTunnelInfo().getRefreshInterval(), "采集通道数据更新任务" + baseProtocolTunnel.getTunnelInfo().getTunnelName(), () -> {
                ((MasterInterface) baseProtocolTunnel).getData();
                return null;
            });
        } else if (baseProtocolTunnel instanceof SlaverInterface) {
            callable = RecurringTaskContainer.getInstance().addRecurringTask(baseProtocolTunnel.getTunnelInfo().getRefreshInterval(), "转发通道数据更新任务" + baseProtocolTunnel.getTunnelInfo().getTunnelName(), () -> {
                ((SlaverInterface) baseProtocolTunnel).updateData2Protocol();
                return null;
            });
        }
        this.tasks.put(baseProtocolTunnel.getTunnelId(), callable);
    }

    @Override
    public String toString() {
        Map<String, JSONObject> str = new HashMap<>();
        this.tunnels.forEach((k, v) ->
                str.put(k, v.getJSONObj())
        );
        return JSON.toJSONString(str);
    }

    /**
     * 静态内部类 为了创建安全的单例
     */
    private static class LazyHolder {
        private static final ProtocolTunnelContainer INSTANCE = new ProtocolTunnelContainer();
    }

    /**
     * 创建一个对通道池内通道状态的检查任务
     */
    class CheckTunnelStatusTask implements Callable {
        @Override
        public Object call() throws Exception {
            TunnelStatus tunnelStatus;
            log.debug("开始对所有的通道状态进行巡检");
            Iterator<String> it = tunnels.keySet().iterator();
            String keyString;
            BaseProtocolTunnel tunnel;
            BaseTunnelInfo tunnelInfo;
            while (it.hasNext()) {
                keyString = it.next();
                tunnel = tunnels.get(keyString);
                tunnelInfo = tunnel.getTunnelInfo();
                tunnelStatus = tunnel.getTunnelStatus();
                if (!TunnelStatus.ABLE_STATUS.contains(tunnelStatus)) {
                    log.warn("发现本该开启的通道:{} 处在未开启成功的状态,重启该通道", tunnelInfo.getTunnelName());
                    try {
                        tunnel.tunnelStop();
                    } catch (DataExchangeException e) {
                        log.error("通道{}关闭时发生异常", tunnelInfo.getTunnelName(), e);
                    }
                    tunnel.refreshTunnelInfo(tunnelInfo).build();
                    try {
                        tunnel.startTunnel();
                    } catch (DataExchangeException e) {
                        log.error("通道{}开启时发生异常", tunnelInfo.getTunnelName(), e);
                    }
                }
            }
            return null;
        }
    }
}